import numpy as np

from semantic_digital_twin.reasoning.predicates import (
    contact,
    visible,
    Above,
    Below,
    LeftOf,
    RightOf,
    Behind,
    InFrontOf,
    is_body_in_region,
    occluding_bodies,
    is_supported_by,
    reachable,
)
from semantic_digital_twin.reasoning.robot_predicates import (
    robot_in_collision,
    robot_holds_body,
    blocking,
    is_body_in_gripper,
)
from semantic_digital_twin.robots.abstract_robot import Camera, ParallelGripper
from semantic_digital_twin.robots.pr2 import PR2
from semantic_digital_twin.testing import *
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.connections import (
    Connection6DoF,
    FixedConnection,
)
from semantic_digital_twin.world_description.geometry import Box, Scale, Color
from semantic_digital_twin.world_description.shape_collection import ShapeCollection
from semantic_digital_twin.world_description.world_entity import Body, Region


@pytest.fixture(scope="function")
def two_block_world():
    def make_body(name: str) -> Body:
        result = Body(name=PrefixedName(name))
        collision = Box(
            scale=Scale(1.0, 1.0, 1.0),
            origin=TransformationMatrix.from_xyz_rpy(reference_frame=result),
        )
        result.collision = ShapeCollection([collision], reference_frame=result)
        return result

    world = World()

    body_1 = make_body("body_1")
    body_2 = make_body("body_2")

    with world.modify_world():
        connection = FixedConnection(
            parent=body_1,
            child=body_2,
            parent_T_connection_expression=TransformationMatrix.from_xyz_rpy(
                z=3, reference_frame=body_1
            ),
        )
        world.add_connection(connection)
    return body_1, body_2


def test_in_contact():
    w = World()

    b1 = Body(name=PrefixedName("b1"))
    collision1 = Box(
        scale=Scale(1.0, 1.0, 1.0),
        origin=TransformationMatrix.from_xyz_rpy(
            0,
            0,
            0.0,
            0,
            0,
            0,
            reference_frame=b1,
        ),
        color=Color(1.0, 0.0, 0.0),
    )
    b1.collision = ShapeCollection([collision1])

    b2 = Body(name=PrefixedName("b2"))
    collision2 = Box(
        scale=Scale(1.0, 1.0, 1.0),
        origin=TransformationMatrix.from_xyz_rpy(
            0.9, 0, 0.0, 0, 0, 0, reference_frame=b2
        ),
        color=Color(0.0, 1.0, 0.0),
    )
    b2.collision = ShapeCollection([collision2])

    b3 = Body(name=PrefixedName("b3"))
    collision3 = Box(
        scale=Scale(1.0, 1.0, 1.0),
        origin=TransformationMatrix.from_xyz_rpy(
            1.8, 0, 0.0, 0, 0, 0, reference_frame=b3
        ),
        color=Color(0.0, 0.0, 1.0),
    )
    b3.collision = ShapeCollection([collision3])

    with w.modify_world():
        w.add_kinematic_structure_entity(b1)
        w.add_kinematic_structure_entity(b2)
        w.add_kinematic_structure_entity(b3)
        w.add_connection(Connection6DoF.create_with_dofs(parent=b1, child=b2, world=w))
        w.add_connection(Connection6DoF.create_with_dofs(parent=b2, child=b3, world=w))
    assert contact(b1, b2)
    assert not contact(b1, b3)
    assert contact(b2, b3)


def test_robot_in_contact(pr2_world: World):
    pr2: PR2 = PR2.from_world(pr2_world)

    body = Body(name=PrefixedName("test_body"))
    collision1 = Box(
        scale=Scale(1.0, 1.0, 1.0),
        origin=TransformationMatrix.from_xyz_rpy(
            z=0.5,
            reference_frame=body,
        ),
        color=Color(1.0, 0.0, 0.0),
    )
    body.collision = ShapeCollection([collision1])

    with pr2_world.modify_world():
        pr2_world.add_connection(
            Connection6DoF.create_with_dofs(
                parent=pr2_world.root, child=body, world=pr2_world
            )
        )

    # Ensure the call runs without raising
    assert robot_in_collision(pr2)

    body.parent_connection.origin = TransformationMatrix.from_xyz_rpy(
        4, 0, 0.5, 0, 0, 0, pr2_world.root
    )
    assert not robot_in_collision(pr2)


def test_get_visible_objects(pr2_world: World):

    pr2: PR2 = PR2.from_world(pr2_world)

    body = Body(name=PrefixedName("test_body"))
    collision1 = Box(
        scale=Scale(1.0, 1.0, 1.0),
        origin=TransformationMatrix.from_xyz_rpy(
            x=2.0,
            z=1.0,
            reference_frame=body,
        ),
        color=Color(1.0, 0.0, 0.0),
    )
    body.collision = ShapeCollection([collision1])

    with pr2_world.modify_world():
        pr2_world.add_connection(
            Connection6DoF.create_with_dofs(
                parent=pr2_world.root, child=body, world=pr2_world
            )
        )

    camera = pr2_world.get_semantic_annotations_by_type(Camera)[0]

    assert visible(camera, body)


def test_occluding_bodies(pr2_world: World):
    pr2: PR2 = PR2.from_world(pr2_world)

    def make_body(name: str) -> Body:
        result = Body(name=PrefixedName(name))
        collision = Box(
            scale=Scale(1.0, 1.0, 1.0),
            origin=TransformationMatrix.from_xyz_rpy(reference_frame=result),
        )
        result.collision = ShapeCollection([collision])
        return result

    obstacle = make_body("obstacle")
    occluded_body = make_body("occluded_body")

    with pr2_world.modify_world():
        root = pr2_world.root
        c1 = FixedConnection(
            parent=root,
            child=obstacle,
            parent_T_connection_expression=TransformationMatrix.from_xyz_rpy(
                reference_frame=root, x=3, z=0.8
            ),
        )
        c2 = FixedConnection(
            parent=root,
            child=occluded_body,
            parent_T_connection_expression=TransformationMatrix.from_xyz_rpy(
                reference_frame=root, x=10, z=0.5
            ),
        )
        pr2_world.add_connection(c1)
        pr2_world.add_connection(c2)

    camera = pr2_world.get_semantic_annotations_by_type(Camera)[0]

    bodies = occluding_bodies(camera, occluded_body)
    assert obstacle in bodies
    assert camera not in bodies
    assert occluded_body not in bodies


def test_above_and_below(two_block_world):
    center, top = two_block_world

    pov = TransformationMatrix.from_xyz_rpy(x=-3)
    assert Above(top, center, pov)()
    assert Below(center, top, pov)()

    pov = TransformationMatrix.from_xyz_rpy(x=3, yaw=np.pi)
    assert Above(top, center, pov)()
    assert Below(center, top, pov)()

    pov = TransformationMatrix.from_xyz_rpy(x=3, roll=np.pi)
    assert Above(center, top, pov)()
    assert Below(top, center, pov)()


def test_left_and_right(two_block_world):
    center, top = two_block_world

    pov = TransformationMatrix.from_xyz_rpy(x=3, roll=np.pi / 2)
    assert LeftOf(top, center, pov)()
    assert RightOf(center, top, pov)()

    pov = TransformationMatrix.from_xyz_rpy(x=3, roll=-np.pi / 2)
    assert RightOf(top, center, pov)()
    assert LeftOf(center, top, pov)()


def test_behind_and_in_front_of(two_block_world):
    center, top = two_block_world

    pov = TransformationMatrix.from_xyz_rpy(z=-5, pitch=np.pi / 2)
    assert Behind(top, center, pov)()
    assert InFrontOf(center, top, pov)()

    pov = TransformationMatrix.from_xyz_rpy(z=5, pitch=-np.pi / 2)
    assert InFrontOf(top, center, pov)()
    assert Behind(center, top, pov)()


def test_body_in_region(two_block_world):
    center, top = two_block_world
    region = Region(name=PrefixedName("test_region"))
    region_box = Box(
        scale=Scale(1.0, 1.0, 1.0),
        origin=TransformationMatrix.from_xyz_rpy(reference_frame=region),
    )
    region.area = ShapeCollection([region_box])

    with center._world.modify_world():
        connection = FixedConnection(
            parent=center,
            child=region,
            parent_T_connection_expression=TransformationMatrix.from_xyz_rpy(
                z=0.5, reference_frame=center
            ),
        )
        center._world.add_connection(connection)
    assert is_body_in_region(center, region) == 0.5
    assert is_body_in_region(top, region) == 0.0


def test_supporting(two_block_world):
    center, top = two_block_world

    with center._world.modify_world():
        top.parent_connection.parent_T_connection_expression = (
            TransformationMatrix.from_xyz_rpy(reference_frame=center, z=1.0)
        )
    assert is_supported_by(top, center)
    assert not is_supported_by(center, top)


def test_is_body_in_gripper(
    pr2_world,
):
    pr2: PR2 = PR2.from_world(pr2_world)

    gripper = pr2_world.get_semantic_annotations_by_type(ParallelGripper)

    left_gripper = (
        gripper[0]
        if LeftOf(gripper[0].root, gripper[1].root, pr2.root.global_pose)()
        else gripper[1]
    )

    # Create test box between fingers
    test_box = Body(name=PrefixedName("test_box"))
    box_collision = Box(
        scale=Scale(0.05, 0.01, 0.05),
        origin=TransformationMatrix.from_xyz_rpy(reference_frame=test_box),
        color=Color(1.0, 0.0, 0.0),
    )
    test_box.collision = ShapeCollection([box_collision])

    # Calculate position between fingers
    finger1_pos = (
        left_gripper.finger.tip.collision.center_of_mass_in_world().to_vector3()
    )
    finger2_pos = (
        left_gripper.thumb.tip.collision.center_of_mass_in_world().to_vector3()
    )
    between_fingers = (finger1_pos + finger2_pos) / 2.0

    # Add box to world
    with pr2_world.modify_world():
        root = pr2_world.root
        connection = Connection6DoF.create_with_dofs(
            parent=root,
            child=test_box,
            world=pr2_world,
        )
        pr2_world.add_connection(connection)
        connection.origin = TransformationMatrix.from_xyz_rpy(
            x=between_fingers[0],
            y=between_fingers[1],
            z=between_fingers[2],
            reference_frame=root,
        )

    assert is_body_in_gripper(test_box, left_gripper) > 0
    assert robot_holds_body(pr2, test_box)
    connection.origin = TransformationMatrix()
    assert is_body_in_gripper(test_box, left_gripper) == 0


def test_reachable(pr2_world):
    pr2: PR2 = PR2.from_world(pr2_world)

    tool_frame_T_reachable_goal = TransformationMatrix.from_xyz_rpy(
        x=-0.2,
        y=0.3,
        reference_frame=pr2.left_arm.manipulator.tool_frame,
    )

    assert reachable(
        tool_frame_T_reachable_goal,
        pr2.left_arm.root,
        pr2.left_arm.manipulator.tool_frame,
    )
    assert not blocking(
        tool_frame_T_reachable_goal,
        pr2.left_arm.root,
        pr2.left_arm.manipulator.tool_frame,
    )
    tool_frame_T_unreachable_goal = TransformationMatrix.from_xyz_rpy(
        x=10, y=10, reference_frame=pr2.left_arm.manipulator.tool_frame
    )
    assert not reachable(
        tool_frame_T_unreachable_goal,
        pr2.left_arm.root,
        pr2.left_arm.manipulator.tool_frame,
    )

    tool_frame_T_rotated_reachable_goal = TransformationMatrix.from_xyz_rpy(
        x=-0.2,
        y=0.3,
        yaw=np.pi / 2,
        reference_frame=pr2.left_arm.manipulator.tool_frame,
    )
    assert reachable(
        tool_frame_T_rotated_reachable_goal,
        pr2.left_arm.root,
        pr2.left_arm.manipulator.tool_frame,
    )

    tool_frame_T_rotated_unreachable_goal = TransformationMatrix.from_xyz_rpy(
        x=-0.2,
        y=0.3,
        yaw=-np.pi / 2,
        reference_frame=pr2.left_arm.manipulator.tool_frame,
    )
    assert not reachable(
        tool_frame_T_rotated_unreachable_goal,
        pr2.left_arm.root,
        pr2.left_arm.manipulator.tool_frame,
    )


def test_blocking(pr2_world):

    pr2: PR2 = PR2.from_world(pr2_world)

    obstacle = Body(name=PrefixedName("obstacle"))
    collision = Box(
        scale=Scale(3.0, 1.0, 1.0),
        origin=TransformationMatrix.from_xyz_rpy(x=1.0, z=0.5),
    )
    obstacle.collision = ShapeCollection([collision])

    with pr2_world.modify_world():
        new_root = Body(name=PrefixedName("new_root"))
        pr2_world.add_connection(
            Connection6DoF.create_with_dofs(
                parent=new_root, child=pr2_world.root, world=pr2_world
            )
        )
        pr2_world.add_connection(
            Connection6DoF.create_with_dofs(
                parent=new_root,
                child=obstacle,
                world=pr2_world,
            )
        )

    assert obstacle not in pr2.bodies
    assert robot_in_collision(pr2)

    tool_frame_T_reachable_goal = TransformationMatrix.from_xyz_rpy(
        x=-0.2,
        y=0.3,
        reference_frame=pr2.left_arm.manipulator.tool_frame,
    )
    assert blocking(
        tool_frame_T_reachable_goal,
        pr2.left_arm.root,
        pr2.left_arm.manipulator.tool_frame,
    )



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from functools import cached_property
from typing import List, Type, Union

from ..reasoning.predicates import LeftOf, RightOf
from ..robots.abstract_robot import Neck, Arm


@dataclass
class HasNeck(ABC):
    """
    Mixin class for robots that have a neck.
    """

    neck: Neck = field(default=None)
    """
    The neck kinematic chain of the robot, if it exists.
    """

    def add_neck(self, neck: Neck):
        """
        Adds a neck kinematic chain to the robot.

        :param neck: The neck kinematic chain to add.
        """
        if not neck.sensors:
            raise ValueError(
                f"Neck kinematic chain {neck.name} must have at least one sensor."
            )
        if self.neck is not None:
            raise ValueError(f"Robot {self.name} already has a neck: {self.neck.name}.")
        self.neck = neck
        self.add_kinematic_chain(neck)


@dataclass
class HasArms(ABC):
    """
    Mixin class for robots that have arms.
    """

    arms: List[Arm] = field(default_factory=list)
    """
    A collection of arms in the robot.
    """

    def add_arm(self, arm: Arm):
        """
        Adds a kinematic chain to the PR2 robot's collection of kinematic chains.
        If the kinematic chain is an arm, it will be added to the left or right arm accordingly.

        :param arm: The kinematic chain to add to the PR2 robot.
        """
        if arm.manipulator is None:
            raise ValueError(f"Arm kinematic chain {arm.name} must have a manipulator.")
        self.arms.append(arm)
        self.add_kinematic_chain(arm)


@dataclass
class SpecifiesLeftRightArm(HasArms, ABC):
    """
    Mixin class for robots that have two arms and can specify which is the left and which is the right arm.
    """

    @cached_property
    def left_arm(self):
        return self._assign_left_right_arms(LeftOf)

    @cached_property
    def right_arm(self):
        return self._assign_left_right_arms(RightOf)

    def _assign_left_right_arms(self, relation: Type[Union[LeftOf, RightOf]]) -> Arm:
        """
        Assigns the left and right arms based on their position relative to the robot's root body.
        :param relation: The relation to use for determining left or right (LeftOf or RightOf).
        :return: The arm that is on the left or right side of the robot.
        """
        assert (
            len(self.arms) == 2
        ), f"Must have exactly two arms to specify left and right arm, but found {len(self.arms)}."
        pov = self.root.global_pose
        first_arm = self.arms[0]
        second_arm = self.arms[1]
        first_arm_chain = list(first_arm.bodies)
        second_arm_chain = list(second_arm.bodies)

        return first_arm if relation(
                first_arm_chain[1],
                second_arm_chain[1],
                pov,
            )() else second_arm



from __future__ import annotations

from abc import abstractmethod, ABC
from dataclasses import dataclass, field

from random_events.utils import SubclassJSONSerializer
from typing_extensions import (
    List,
    Dict,
    Any,
    Self,
    Callable,
    TYPE_CHECKING,
)

from .degree_of_freedom import DegreeOfFreedom
from .world_entity import KinematicStructureEntity, SemanticAnnotation, Connection
from ..datastructures.prefixed_name import PrefixedName

if TYPE_CHECKING:
    from ..world import World


@dataclass
class UnknownWorldModification(Exception):
    """
    Raised when an unknown world modification is attempted.
    """

    call: Callable
    kwargs: Dict[str, Any]

    def __post_init__(self):
        super().__init__(
            " Make sure that world modifications are atomic and that every atomic modification is "
            "represented by exactly one subclass of WorldModelModification."
            "This module might be incomplete, you can help by expanding it."
        )


@dataclass
class WorldModelModification(SubclassJSONSerializer, ABC):
    """
    A record of a modification to the model (structure) of the world.
    This includes add/remove body and add/remove connection.

    All modifications are compared via the names of the objects they reference.

    This class is referenced by the `atomic_world_modification` decorator and should be used for a method that
    applies such a modification to the world.
    """

    @abstractmethod
    def apply(self, world: World):
        """
        Apply this change to the given world.

        :param world: The world to modify.
        """

    @classmethod
    @abstractmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]) -> Self:
        """
        Factory to construct this change from the kwargs of its corresponding method in World decorated with
        `atomic_world_modification(modification=cls)`.

        :param kwargs: The kwargs of the function call.
        :return: A new instance.
        """
        raise NotImplementedError


@dataclass
class AddKinematicStructureEntityModification(WorldModelModification):
    """
    Addition of a body to the world.
    """

    kinematic_structure_entity: KinematicStructureEntity
    """
    The body that was added.
    """

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]):
        return cls(kwargs["kinematic_structure_entity"])

    def apply(self, world: World):
        world.add_kinematic_structure_entity(self.kinematic_structure_entity)

    def to_json(self):
        return {**super().to_json(), "body": self.kinematic_structure_entity.to_json()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            kinematic_structure_entity=KinematicStructureEntity.from_json(data["body"])
        )

    def __eq__(self, other: Any) -> bool:
        return (
            self.kinematic_structure_entity.name
            == other.kinematic_structure_entity.name
        )


@dataclass
class RemoveBodyModification(WorldModelModification):
    """
    Removal of a body from the world.
    """

    body_name: PrefixedName
    """
    The name of the body that was removed.
    """

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]):
        return cls(kwargs["kinematic_structure_entity"].name)

    def apply(self, world: World):
        world.remove_kinematic_structure_entity(
            world.get_kinematic_structure_entity_by_name(self.body_name)
        )

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "body_name": self.body_name.to_json()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(body_name=PrefixedName.from_json(data["body_name"]))


@dataclass
class AddConnectionModification(WorldModelModification):
    """
    Addition of a connection to the world.
    """

    connection: Connection
    """
    The connection that was added.
    """

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]):
        return cls(kwargs["connection"])

    def apply(self, world: World):
        self.connection.parent = world.get_kinematic_structure_entity_by_name(
            self.connection.parent.name
        )
        self.connection.child = world.get_kinematic_structure_entity_by_name(
            self.connection.child.name
        )
        world.add_connection(self.connection)

    def to_json(self):
        return {
            **super().to_json(),
            "connection": self.connection.to_json(),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(connection=Connection.from_json(data["connection"]))

    def __eq__(self, other):
        return (
            isinstance(other, AddConnectionModification)
            and self.connection.name == other.connection.name
        )


@dataclass
class RemoveConnectionModification(WorldModelModification):
    """
    Removal of a connection from the world.
    """

    connection_name: PrefixedName
    """
    The name of the connection that was removed.
    """

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]):
        return cls(kwargs["connection"].name)

    def apply(self, world: World):
        world._remove_connection(world.get_connection_by_name(self.connection_name))

    def to_json(self):
        return {
            **super().to_json(),
            "connection_name": self.connection_name.to_json(),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(connection_name=PrefixedName.from_json(data["connection_name"]))


@dataclass
class AddDegreeOfFreedomModification(WorldModelModification):
    """
    Addition of a degree of freedom to the world.
    """

    dof: DegreeOfFreedom
    """
    The degree of freedom that was added.
    """

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]):
        return cls(dof=kwargs["dof"])

    def apply(self, world: World):
        world.add_degree_of_freedom(self.dof)

    def to_json(self):
        return {
            **super().to_json(),
            "dof": self.dof.to_json(),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(dof=DegreeOfFreedom.from_json(data["dof"]))

    def __eq__(self, other):
        return self.dof.name == other.dof.name


@dataclass
class RemoveDegreeOfFreedomModification(WorldModelModification):
    dof_name: PrefixedName

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]):
        return cls(dof_name=kwargs["dof"].name)

    def apply(self, world: World):
        world.remove_degree_of_freedom(
            world.get_degree_of_freedom_by_name(self.dof_name)
        )

    def to_json(self):
        return {
            **super().to_json(),
            "dof": self.dof_name.to_json(),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(dof_name=PrefixedName.from_json(data["dof"]))


@dataclass
class AddSemanticAnnotationModification(WorldModelModification):
    semantic_annotation: SemanticAnnotation

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]):
        return cls(semantic_annotation=kwargs["semantic_annotation"])

    def apply(self, world: World):
        world.add_semantic_annotation(self.semantic_annotation)

    def to_json(self):
        return {
            **super().to_json(),
            "semantic_annotation": self.semantic_annotation.to_json(),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            semantic_annotation=SemanticAnnotation.from_json(
                data["semantic_annotation"]
            )
        )


@dataclass
class RemoveSemanticAnnotationModification(WorldModelModification):
    semantic_annotation: SemanticAnnotation

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]):
        return cls(semantic_annotation=kwargs["semantic_annotation"])

    def apply(self, world: World):
        world.remove_semantic_annotation(self.semantic_annotation)

    def to_json(self):
        return {
            **super().to_json(),
            "semantic_annotation": self.semantic_annotation.to_json(),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            semantic_annotation=SemanticAnnotation.from_json(
                data["semantic_annotation"]
            )
        )


@dataclass
class WorldModelModificationBlock(SubclassJSONSerializer):
    """
    A sequence of WorldModelModifications that were applied to the world within one `with world.modify_world()` context.
    """

    modifications: List[WorldModelModification] = field(default_factory=list)
    """
    The list of modifications to apply to the world.
    """

    def apply(self, world: World):
        with world.modify_world():
            for modification in self.modifications:
                modification.apply(world)

    def to_json(self):
        return {
            **super().to_json(),
            "modifications": [m.to_json() for m in self.modifications],
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls([WorldModelModification.from_json(d) for d in data["modifications"]])

    def __iter__(self):
        return iter(self.modifications)

    def __getitem__(self, item):
        return self.modifications[item]

    def __len__(self):
        return len(self.modifications)

    def append(self, modification: WorldModelModification):
        self.modifications.append(modification)


@dataclass
class SetDofHasHardwareInterface(WorldModelModification):
    degree_of_freedom_names: List[PrefixedName]
    value: bool

    def apply(self, world: World):
        for dof_name in self.degree_of_freedom_names:
            world.get_degree_of_freedom_by_name(dof_name).has_hardware_interface = (
                self.value
            )

    @classmethod
    def from_kwargs(cls, kwargs: Dict[str, Any]) -> Self:
        dofs = kwargs["dofs"]
        degree_of_freedom_names = [dof.name for dof in dofs]
        return cls(
            degree_of_freedom_names=degree_of_freedom_names, value=kwargs["value"]
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "degree_of_freedom_names": [
                dof.to_json() for dof in self.degree_of_freedom_names
            ],
            "value": self.value,
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            degree_of_freedom_names=[
                PrefixedName.from_json(dof) for dof in data["degree_of_freedom_names"]
            ],
            value=data["value"],
        )



from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np

from probabilistic_model.probabilistic_circuit.rx.helper import uniform_measure_of_event
from typing_extensions import List

from ..world_description.shape_collection import BoundingBoxCollection
from ..spatial_types import Point3
from ..datastructures.variables import SpatialVariables
from ..world_description.world_entity import SemanticAnnotation, Body, Region


@dataclass(eq=False)
class HasDrawers:
    """
    A mixin class for semantic annotations that have drawers.
    """

    drawers: List[Drawer] = field(default_factory=list, hash=False)


@dataclass(eq=False)
class HasDoors:
    """
    A mixin class for semantic annotations that have doors.
    """

    doors: List[Door] = field(default_factory=list, hash=False)


@dataclass(eq=False)
class Handle(SemanticAnnotation):
    body: Body


@dataclass(eq=False)
class Container(SemanticAnnotation):
    body: Body


@dataclass(eq=False)
class Fridge(SemanticAnnotation):
    """
    A semantic annotation representing a fridge that has a door and a body.
    """

    body: Body
    door: Door


@dataclass(eq=False)
class Table(SemanticAnnotation):
    """
    A semantic annotation that represents a table.
    """

    top: Body
    """
    The body that represents the table's top surface.
    """

    def points_on_table(self, amount: int = 100) -> List[Point3]:
        """
        Get points that are on the table.

        :amount: The number of points to return.
        :returns: A list of points that are on the table.
        """
        area_of_table = BoundingBoxCollection.from_shapes(self.top.collision)
        event = area_of_table.event
        p = uniform_measure_of_event(event)
        p = p.marginal(SpatialVariables.xy)
        samples = p.sample(amount)
        z_coordinate = np.full(
            (amount, 1), max([b.max_z for b in area_of_table]) + 0.01
        )
        samples = np.concatenate((samples, z_coordinate), axis=1)
        return [Point3(*s, reference_frame=self.top) for s in samples]


################################


@dataclass(eq=False)
class Components(SemanticAnnotation): ...


@dataclass(eq=False)
class Furniture(SemanticAnnotation): ...


@dataclass(eq=False)
class SupportingSurface(SemanticAnnotation):
    """
    A semantic annotation that represents a supporting surface.
    """

    region: Region
    """
    The region that represents the supporting surface.
    """


#################### subclasses von Components


@dataclass(eq=False)
class EntryWay(Components):
    body: Body


@dataclass(eq=False)
class Door(EntryWay):
    handle: Handle


@dataclass(eq=False)
class DoubleDoor(EntryWay):
    left_door: Door
    right_door: Door


@dataclass(eq=False)
class Drawer(Components):
    container: Container
    handle: Handle


############################### subclasses to Furniture
@dataclass(eq=False)
class Cabinet(Furniture):
    container: Container
    drawers: List[Drawer] = field(default_factory=list, hash=False)
    doors: List[Door] = field(default_factory=list)


@dataclass(eq=False)
class Dresser(Furniture):
    container: Container
    drawers: List[Drawer] = field(default_factory=list, hash=False)
    doors: List[Door] = field(default_factory=list)


@dataclass(eq=False)
class Cupboard(Furniture):
    container: Container
    doors: List[Door] = field(default_factory=list)


@dataclass(eq=False)
class Wardrobe(Furniture):
    container: Container
    drawers: List[Drawer] = field(default_factory=list, hash=False)
    doors: List[Door] = field(default_factory=list)


class Floor(SupportingSurface): ...


@dataclass(eq=False)
class Room(SemanticAnnotation):
    """
    A semantic annotation that represents a closed area with a specific purpose
    """

    floor: Floor
    """
    The room's floor.
    """


@dataclass(eq=False)
class Wall(SemanticAnnotation):
    body: Body
    doors: List[Door] = field(default_factory=list)



from __future__ import annotations

import os
import re
import time

import tqdm
from krrood.ormatic.dao import to_dao
from krrood.ormatic.utils import drop_database
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from typing_extensions import TYPE_CHECKING

from semantic_digital_twin.adapters.fbx import FBXParser
from semantic_digital_twin.adapters.procthor.procthor_pipelines import (
    dresser_factory_from_body,
)
from semantic_digital_twin.orm.ormatic_interface import *
from semantic_digital_twin.pipeline.pipeline import (
    Pipeline,
    BodyFilter,
    BodyFactoryReplace,
    CenterLocalGeometryAndPreserveWorldPose,
)

if TYPE_CHECKING:
    from semantic_digital_twin.world import World


def remove_root_and_move_children_into_new_worlds(world: World) -> List[World]:
    """
    Remove the root of the given world and move its children into new worlds.
    Each child that has a parent with "grp" in its name and does not have "grp" in its own name
    will be moved to a new world, (sometimes groups are nested).
    The new world's name will be set to the child's name.

    :param world: The World object to process. This world will be unusable after this operation.
    :return: List of new World objects created from the root's children.
    """
    root_children = [
        entity
        for entity in world.kinematic_structure_entities
        if entity.parent_kinematic_structure_entity
        and "grp" in entity.parent_kinematic_structure_entity.name.name
        and "grp" not in entity.name.name
    ]

    with world.modify_world():

        worlds = [world.copy_subgraph_to_new_world(child) for child in root_children]
        for world in worlds:
            world.name = world.root.name.name

    return worlds


def replace_dresser_meshes_with_factories(
    worlds: List[World], dresser_pattern: re.Pattern[str]
) -> List[World]:
    """
    Replace dresser meshes in the given worlds with dresser factories.
    A dresser is identified by its name matching the given dresser_pattern regex.

    :param worlds: List of World objects to process.
    :param dresser_pattern: A compiled regex pattern to identify dresser bodies.
    :return: List of World objects with dresser meshes replaced by factories.
    """
    procthor_factory_replace_pipeline = Pipeline(
        [
            BodyFactoryReplace(
                body_condition=lambda b: bool(dresser_pattern.fullmatch(b.name.name))
                and not (
                    "drawer" in b.name.name.lower() or "door" in b.name.name.lower()
                ),
                factory_creator=dresser_factory_from_body,
            )
        ]
    )
    worlds = [procthor_factory_replace_pipeline.apply(w) for w in worlds]
    return worlds


def parse_fbx_file_to_world_mapping_daos(fbx_file_path: str) -> List[WorldMappingDAO]:
    """
    Parse a Procthor FBX file path and return a list of WorldMappingDAO objects.
    """
    dresser_pattern = re.compile(r"^.*dresser_(?!drawer\b).*$", re.IGNORECASE)

    pipeline = Pipeline(
        [
            CenterLocalGeometryAndPreserveWorldPose(),
            BodyFilter(lambda x: not x.name.name.startswith("PS_")),
            BodyFilter(lambda x: not x.name.name.endswith("slice")),
        ]
    )

    parser = FBXParser(fbx_file_path)
    world = parser.parse()
    world = pipeline.apply(world)

    worlds = remove_root_and_move_children_into_new_worlds(world)

    worlds = replace_dresser_meshes_with_factories(worlds, dresser_pattern)

    return [to_dao(world) for world in worlds]


def parse_procthor_files_and_save_to_database(
    fbx_file_pattern: re.Pattern[str] = re.compile(r".*_grp\.fbx$", re.IGNORECASE),
    drop_existing_database: bool = True,
):
    """
    Parse all Procthor FBX files and store the resulting WorldMappingDAO objects in a database.
    Currently, only grp files are parsed, and some files and names are excluded.
    TODO: Ensure all relevant files, even those not inside a grp, are parsed.
    """
    semantic_digital_twin_database_uri = os.environ.get(
        "semantic_digital_twin_DATABASE_URI"
    )
    assert (
        semantic_digital_twin_database_uri is not None
    ), "Please set the semantic_digital_twin_DATABASE_URI environment variable."

    procthor_root = os.path.join(os.path.expanduser("~"), "ai2thor")
    # procthor_root = os.path.join(os.path.expanduser("~"), "work", "ai2thor")

    files = []
    for root, dirs, filenames in os.walk(procthor_root):
        for filename in filenames:
            files.append(os.path.join(root, filename))

    assert (
        len(files) > 0
    ), f"No files found in {procthor_root}, please set the correct path to the ai2thor directory."

    excluded_words = [
        "FirstPersonCharacter",
        "SourceFiles_Procedural",
        "RobotArmTest",
        "_shards_",
    ]

    fbx_files = [
        f
        for f in files
        if not any([e in f for e in excluded_words]) and fbx_file_pattern.fullmatch(f)
    ]
    # Create database engine and session
    engine = create_engine(f"mysql+pymysql://{semantic_digital_twin_database_uri}")
    session = Session(engine)

    if drop_existing_database:
        drop_database(engine)
        Base.metadata.create_all(engine)

    start_time = time.time_ns()

    dao_names = []
    daos = []

    for fbx_file in tqdm.tqdm(fbx_files):
        for dao in parse_fbx_file_to_world_mapping_daos(fbx_file):
            # Some item names (for example "bowl_19") were used for multiple items. For now the solution is to just
            # skip duplicate names.
            if dao.name not in dao_names:
                dao_names.append(dao.name)
                daos.append(dao)

    session.add_all(daos)
    session.commit()
    print(
        f"Parsing {len(fbx_files)} files took {(time.time_ns() - start_time) / 1e9} seconds"
    )


if __name__ == "__main__":
    parse_procthor_files_and_save_to_database()



import atexit
import threading
import time
from dataclasses import dataclass

import numpy as np
import rclpy.node

from .. import logger
from ..callbacks.callback import StateChangeCallback

try:
    from builtin_interfaces.msg import Duration
    from geometry_msgs.msg import Vector3, Point, Quaternion, Pose
    from std_msgs.msg import ColorRGBA
    from visualization_msgs.msg import Marker, MarkerArray
    from geometry_msgs.msg import Vector3, Point, PoseStamped, Quaternion, Pose
except ImportError as e:
    logger.warning(
        f"Could not import ros messages, viz marker will not be available: {e}"
    )

from scipy.spatial.transform import Rotation

from ..world_description.geometry import (
    FileMesh,
    Box,
    Sphere,
    Cylinder,
    TriangleMesh,
)
from ..world import World


@dataclass
class VizMarkerPublisher(StateChangeCallback):
    """
    Publishes an Array of visualization marker which represent the situation in the World
    """

    node: rclpy.node.Node
    """
    The ROS2 node that will be used to publish the visualization marker.
    """

    topic_name: str = "/semworld/viz_marker"
    """
    The name of the topic to which the Visualization Marker should be published.
    """

    reference_frame: str = "map"
    """
    The reference frame of the visualization marker.
    """

    def __post_init__(self):
        """
        Initializes the publisher and registers the callback to the world.
        """
        super().__post_init__()
        self.pub = self.node.create_publisher(MarkerArray, self.topic_name, 10)
        time.sleep(0.2)
        self.notify()

    def _notify(self):
        """
        Publishes the Marker Array on world changes.
        """
        marker_array = self._make_marker_array()
        self.pub.publish(marker_array)

    def _make_marker_array(self) -> MarkerArray:
        """
        Creates the Marker Array to be published. There is one Marker for link for each object in the Array, each Object
        creates a name space in the visualization Marker. The type of Visualization Marker is decided by the collision
        tag of the URDF.

        :return: An Array of Visualization Marker
        """
        marker_array = MarkerArray()
        for body in self.world.bodies:
            for i, collision in enumerate(body.collision):
                msg = Marker()
                msg.header.frame_id = self.reference_frame
                msg.ns = body.name.name
                msg.id = i
                msg.action = Marker.ADD
                msg.pose = self.transform_to_pose(
                    (
                        self.world.compute_forward_kinematics(self.world.root, body)
                        @ collision.origin
                    ).to_np()
                )
                msg.color = ColorRGBA(
                    r=float(collision.color.R),
                    g=float(collision.color.G),
                    b=float(collision.color.B),
                    a=float(collision.color.A),
                )
                msg.lifetime = Duration(sec=0)

                if isinstance(collision, FileMesh):
                    msg.type = Marker.MESH_RESOURCE
                    msg.mesh_resource = "file://" + collision.filename
                    msg.scale = Vector3(
                        x=float(collision.scale.x),
                        y=float(collision.scale.y),
                        z=float(collision.scale.z),
                    )
                    msg.mesh_use_embedded_materials = True
                elif isinstance(collision, TriangleMesh):
                    f = collision.file
                    msg.type = Marker.MESH_RESOURCE
                    msg.mesh_resource = "file://" + f.name
                    msg.scale = Vector3(
                        x=float(collision.scale.x),
                        y=float(collision.scale.y),
                        z=float(collision.scale.z),
                    )
                    msg.mesh_use_embedded_materials = True
                elif isinstance(collision, Cylinder):
                    msg.type = Marker.CYLINDER
                    msg.scale = Vector3(
                        x=float(collision.width),
                        y=float(collision.width),
                        z=float(collision.height),
                    )
                elif isinstance(collision, Box):
                    msg.type = Marker.CUBE
                    msg.scale = Vector3(
                        x=float(collision.scale.x),
                        y=float(collision.scale.y),
                        z=float(collision.scale.z),
                    )
                elif isinstance(collision, Sphere):
                    msg.type = Marker.SPHERE
                    msg.scale = Vector3(
                        x=float(collision.radius * 2),
                        y=float(collision.radius * 2),
                        z=float(collision.radius * 2),
                    )

                marker_array.markers.append(msg)
        return marker_array

    @staticmethod
    def transform_to_pose(transform: np.ndarray) -> Pose:
        """
        Converts a 4x4 transformation matrix to a PoseStamped message.

        :param transform: The transformation matrix to convert.
        :return: A PoseStamped message.
        """
        pose = Pose()
        pose.position = Point(**dict(zip(["x", "y", "z"], transform[:3, 3])))
        pose.orientation = Quaternion(
            **dict(
                zip(
                    ["x", "y", "z", "w"],
                    Rotation.from_matrix(transform[:3, :3]).as_quat(),
                )
            )
        )
        return pose



import numpy as np
from ..datastructures.types import AnyMatrix4x4, NpMatrix4x4


def inverse_frame(f1_T_f2: NpMatrix4x4) -> NpMatrix4x4:
    """
    :param f1_T_f2: 4x4 Matrix
    :return: f2_T_f1
    """
    R = f1_T_f2[:3, :3]
    t = f1_T_f2[:3, 3]
    Rt = R.T
    f2_T_f1 = np.empty((4, 4), dtype=f1_T_f2.dtype)
    f2_T_f1[:3, :3] = Rt
    f2_T_f1[:3, 3] = -Rt @ t
    f2_T_f1[3] = [0, 0, 0, 1]
    return f2_T_f1



import json
import os
import unittest
from dataclasses import asdict
from time import sleep

import numpy as np
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from semantic_digital_twin.adapters.procthor.procthor_parser import (
    ProcTHORParser,
    ProcthorRoom,
    unity_to_semantic_digital_twin_transform,
    ProcthorDoor,
    ProcthorWall,
    ProcthorObject,
)
from semantic_digital_twin.spatial_types.spatial_types import TransformationMatrix
from semantic_digital_twin.utils import get_semantic_digital_twin_directory_root
from semantic_digital_twin.world_description.geometry import Scale
from semantic_digital_twin.world_description.world_entity import Region


class ProcTHORTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        json_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "..",
            "..",
            "resources",
            "procthor_json",
        )
        cls.file_path = os.path.join(json_dir, "house_test.json")
        with open(cls.file_path) as f:
            cls.house_json = json.load(f)

    def test_unity_to_semantic_digital_twin_transform_identity_matrix(self):
        m = np.eye(4)
        result = unity_to_semantic_digital_twin_transform(TransformationMatrix(data=m))
        np.testing.assert_allclose(result.to_np(), np.eye(4), rtol=1e-6, atol=1e-6)

    def test_unity_to_semantic_digital_twin_transform_translation_along_x(self):
        """Unity +X should map to semantic –Y (because of reflection)."""
        m = np.eye(4)
        m[0, 3] = 1.0
        result = unity_to_semantic_digital_twin_transform(TransformationMatrix(data=m))
        self.assertAlmostEqual(result.to_position().to_np()[1], -1.0)
        np.testing.assert_allclose(
            result.to_rotation_matrix().to_np()[:3, :3], np.eye(3)
        )

    def test_unity_to_semantic_digital_twin_transform_translation_along_z(self):
        """Unity +Z should map to semantic +X."""
        m = np.eye(4)
        m[2, 3] = 2.0
        result = unity_to_semantic_digital_twin_transform(TransformationMatrix(data=m))
        self.assertAlmostEqual(result.to_position().to_np()[0], 2.0, places=6)
        np.testing.assert_allclose(
            result.to_rotation_matrix().to_np()[:3, :3], np.eye(3)
        )

    def test_unity_to_semantic_digital_twin_transform_rotation_y_90_degrees(self):
        """Unity +90° about Y should become –90° about Z in semantic frame."""
        theta = np.pi / 2
        m = np.eye(4)
        m[:3, :3] = np.array(
            [
                [np.cos(theta), 0, np.sin(theta)],
                [0, 1, 0],
                [-np.sin(theta), 0, np.cos(theta)],
            ]
        )
        result = unity_to_semantic_digital_twin_transform(TransformationMatrix(data=m))

        expected = np.eye(4)
        expected[:3, :3] = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])
        np.testing.assert_allclose(result.to_np(), expected, rtol=1e-6, atol=1e-6)

    def test_room_centered_polytope(self):
        room = self.house_json["rooms"][0]
        procthor_room = ProcthorRoom(room_dict=room)
        point_array = np.asarray(
            [point.to_np()[:3] for point in procthor_room.centered_polytope]
        )
        self.assertTrue(np.allclose(point_array.mean(axis=0), np.zeros(3), atol=1e-6))

    def test_world_T_room(self):
        room = self.house_json["rooms"][0]
        procthor_room = ProcthorRoom(room_dict=room)
        np.testing.assert_array_equal(
            procthor_room.world_T_room.to_rotation_matrix().to_np(), np.eye(4)
        )
        np.testing.assert_array_equal(
            procthor_room.world_T_room.to_translation().to_np()[:3, 3],
            np.array([1.5, -1.5, 0]),
        )

    def test_room_get_world(self):
        room = self.house_json["rooms"][0]
        procthor_room = ProcthorRoom(room_dict=room)
        world = procthor_room.get_world()
        self.assertEqual(world.root.name.name, "Kitchen_4")

        region = world.get_kinematic_structure_entity_by_name("Kitchen_4_region")
        self.assertIsInstance(region, Region)
        region_area_mesh = region.area[0]
        self.assertAlmostEqual(region_area_mesh.mesh.area, 18.06)

    def test_door_polygon(self):
        door = self.house_json["doors"][0]
        procthor_door = ProcthorDoor(
            door_dict=door, parent_wall_width=0.05, thickness=0.03
        )

        bounds = (
            procthor_door.x_min,
            procthor_door.x_max,
            procthor_door.y_min,
            procthor_door.y_max,
        )
        expected_bounds = (0.25, 2.25, 0.0, 2.1)
        np.testing.assert_allclose(bounds, expected_bounds, rtol=1e-6, atol=1e-6)

    def test_door_scale(self):
        door = self.house_json["doors"][0]
        procthor_door = ProcthorDoor(
            door_dict=door, parent_wall_width=0.05, thickness=0.03
        )

        scale = procthor_door.scale
        expected_scale = Scale(0.03, 2.0, 2.1)
        self.assertEqual(expected_scale, scale)

    def test_wall_T_door(self):
        door = self.house_json["doors"][0]
        procthor_door = ProcthorDoor(
            door_dict=door, parent_wall_width=0.05, thickness=0.03
        )

        expected_rotation = np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, -1.225],
                [0, 0, 1, 1.05],
                [0, 0, 0, 1],
            ]
        )
        np.testing.assert_allclose(
            procthor_door.wall_T_door.to_np(),
            expected_rotation,
            rtol=1e-6,
            atol=1e-6,
        )

    def test_door_get_factory(self):
        door = self.house_json["doors"][0]
        procthor_door = ProcthorDoor(
            door_dict=door, parent_wall_width=0.05, thickness=0.03
        )

        door_factory = procthor_door.get_factory()

        self.assertEqual(door_factory.name.name, "Doorway_Double_7_room1_room4")
        self.assertEqual(len(door_factory.door_factories), 2)
        self.assertEqual(len(door_factory.door_transforms), 2)
        self.assertEqual(door_factory.door_factories[0].scale, Scale(0.03, 1.0, 2.1))

    def test_wall_creation(self):
        parser = ProcTHORParser(self.file_path, None)
        doors = self.house_json["doors"]
        walls = self.house_json["walls"]

        procthor_wall = ProcthorWall(wall_dicts=walls, door_dicts=doors)

        procthor_walls_from_door, used_wall_ids = parser._build_procthor_wall_from_door(
            walls, doors
        )

        self.assertEqual(procthor_walls_from_door, [procthor_wall])

        procthor_walls_from_polygon = parser._build_procthor_wall_from_polygon(walls)[0]

        self.assertFalse(procthor_walls_from_polygon.door_dicts)

        self.assertEqual(
            asdict(procthor_walls_from_door[0], dict_factory=dict),
            {
                **asdict(procthor_walls_from_polygon, dict_factory=dict),
                "door_dicts": procthor_walls_from_door[0].door_dicts,
            },
        )

    def test_wall_coords(self):
        doors = self.house_json["doors"]
        walls = self.house_json["walls"]

        procthor_wall = ProcthorWall(wall_dicts=walls, door_dicts=doors)

        x_coords, y_coords, z_coords = (
            procthor_wall.x_coords,
            procthor_wall.y_coords,
            procthor_wall.z_coords,
        )

        np.testing.assert_array_equal(x_coords, [3.5])
        np.testing.assert_array_equal(y_coords, [0.0, 4.2])
        np.testing.assert_array_equal(z_coords, [3.6, 0.0])
        self.assertEqual(procthor_wall.delta_x, 0.0)
        self.assertEqual(procthor_wall.delta_z, 3.6)

    def test_wall_scale(self):
        doors = self.house_json["doors"]
        walls = self.house_json["walls"]

        procthor_wall = ProcthorWall(
            wall_dicts=walls, door_dicts=doors, wall_thickness=0.05
        )

        self.assertEqual(procthor_wall.scale, Scale(0.05, 3.6, 4.2))

    def test_world_T_wall(self):
        doors = self.house_json["doors"]
        walls = self.house_json["walls"]

        procthor_wall = ProcthorWall(
            wall_dicts=walls, door_dicts=doors, wall_thickness=0.05
        )

        expected_translation = np.array(
            [
                [0, 1, 0, 1.8],
                [-1, 0, 0, -3.5],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
            ]
        )
        np.testing.assert_allclose(
            procthor_wall.world_T_wall.to_np(),
            expected_translation,
            rtol=1e-6,
            atol=1e-6,
        )

    @unittest.skip("Requires Database, TBD")
    def test_world_T_obj(self):
        objects = self.house_json["objects"][0]

        semantic_digital_twin_database_uri = os.environ.get(
            "semantic_digital_twin_DATABASE_URI"
        )

        # Create database engine and session
        engine = create_engine(f"mysql+pymysql://{semantic_digital_twin_database_uri}")
        session = Session(engine)

        procthor_object = ProcthorObject(object_dict=objects, session=session)

        desired = np.array(
            [
                [-1.0, 0.0, 0.0, 6.7],
                [0.0, -1.0, 0.0, -6.0],
                [0.0, 0.0, 1.0, 0.5],
                [0.0, 0.0, 0.0, 1.0],
            ]
        )

        np.testing.assert_almost_equal(
            procthor_object.world_T_obj.to_np(),
            desired,
            decimal=6,
        )

    @unittest.skip("Requires Database, TBD")
    def test_object_get_world(self):
        objects = self.house_json["objects"][0]

        semantic_digital_twin_database_uri = os.environ.get(
            "semantic_digital_twin_DATABASE_URI"
        )

        # Create database engine and session
        engine = create_engine(f"mysql+pymysql://{semantic_digital_twin_database_uri}")
        session = Session(engine)

        procthor_object = ProcthorObject(object_dict=objects, session=session)
        world = procthor_object.get_world()

        ...

    def test_parse_full_world(self):
        world = ProcTHORParser(
            os.path.join(
                get_semantic_digital_twin_directory_root(os.getcwd()),
                "resources",
                "procthor_json",
                "house_987654321.json",
            )
        ).parse()

        assert world is not None


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from ctypes import c_int
from dataclasses import dataclass, field
from enum import Enum
from typing_extensions import Dict, TYPE_CHECKING, List, Tuple

import daqp
import numpy as np

from ..world_description.connections import ActiveConnection, PassiveConnection
from ..world_description.degree_of_freedom import DegreeOfFreedom
from ..spatial_types import spatial_types as cas

if TYPE_CHECKING:
    from ..world import World
    from ..world_description.world_entity import KinematicStructureEntity

_large_value = np.inf
"""
Used as bounds for slack variables. 
Only needs to be changed when a different QP solver is used, as some can't handle inf.
"""


class IKSolverException(Exception):
    pass


class UnreachableException(IKSolverException):
    iterations: int
    """
    After how many iterations the solver converged.
    """

    def __init__(self, iterations: int):
        self.iterations = iterations
        super().__init__(
            f"Converged after {self.iterations}, but target pose not reached."
        )


class MaxIterationsException(IKSolverException):
    iterations: int
    """
    After how many iterations the solver did not converge.
    """

    def __init__(self, iterations: int):
        super().__init__(f"Failed to converge in {iterations} iterations.")


class DAQPSolverExitFlag(Enum):
    """
    Exit flags for the DAQP solver.
    """

    SOFT_OPTIMAL = (2, "Soft optimal")
    OPTIMAL = (1, "Optimal")
    INFEASIBLE = (-1, "Infeasible")
    CYCLING_DETECTED = (-2, "Cycling detected")
    UNBOUNDED_PROBLEM = (-3, "Unbounded problem")
    ITERATION_LIMIT_REACHED = (-4, "Iteration limit reached")
    NONCONVEX_PROBLEM = (-5, "Nonconvex problem")
    INITIAL_WORKING_SET_OVERDETERMINED = (-6, "Initial working set overdetermined")

    def __init__(self, code, description):
        self.code = code
        self.description = description

    @classmethod
    def from_code(cls, code):
        for flag in cls:
            if flag.code == code:
                return flag
        raise ValueError(f"Unknown exit flag code: {code}")


class QPSolverException(IKSolverException):
    def __init__(self, exit_flag_code):
        self.exit_flag = DAQPSolverExitFlag.from_code(exit_flag_code)
        super().__init__(
            f"QP solver failed with exit flag: {self.exit_flag.description}"
        )


@dataclass
class InverseKinematicsSolver:
    """
    Quadratic Programming-based Inverse Kinematics solver.

    This class handles the setup and solving of inverse kinematics problems
    using quadratic programming optimization.
    General idea:
    min_{dof_v, slack} low_weight * (dof_v1² + ... + dof_vn²) + high_weight * (slack_trans_x² + ... + slack_rot_z²)
    subject to
        position_error = J_translation * dof_v + translation slack
        rotation_error = J_rotation * dof_v + rotation slack
    where
        dof_v = velocity of dofs, e.g., joint velocity
        slack = auxiliary variables used to make constraints violable, in this case the remaining error that can not be achieved with the velocity.
        J_translation is the jacobian of the forward kinematics position.
        J_rotation is the jacobian of the forward kinematics rotation.
        position_error = target_position - current_position
        rotation_error = target_rotation - current_rotation

    The slack variables ensure that the QP stays solvable, even if the target velocity cannot be reached.
    Slack variables have a high weight, such that the solver only violates the constraints, if necessary.
    If we solve this QP at the current state, we get a velocity for the dofs, apply this to the current state with a dt and solve again.
    dof_v and slack below threshold:
        target reached
    dof_v below threshold, slack above threshold:
        target unreachable
    dof_v above threshold, slack below threshold:
        almost done
    dof_v above threshold, slack above threshold:
        neither close to the target, nor converged
    """

    world: World
    """
    Backreference to semantic digital twin.
    """

    iterations: int = field(default=-1, init=False)
    """
    The current iteration of the solver.
    """

    _convergence_velocity_tolerance: float = field(default=1e-4)
    """
    If the velocity of the active DOFs is below this threshold, the solver is considered to have converged.
    Unit depends on the DOF, e.g. rad/s for revolute joints or m/s for prismatic joints.
    """

    _convergence_slack_tolerance: float = field(default=1e-3)
    """
    The slack variables describe how much the target is violated. 
    If all slack variables are below this threshold, the solver found a solution.
    Unit is m for the position target or rad for the orientation target.
    """

    def solve(
        self,
        root: KinematicStructureEntity,
        tip: KinematicStructureEntity,
        target: cas.TransformationMatrix,
        dt: float = 0.05,
        max_iterations: int = 200,
        translation_velocity: float = 0.2,
        rotation_velocity: float = 0.2,
    ) -> Dict[DegreeOfFreedom, float]:
        """
        Solve inverse kinematics problem.

        :param root: Root body of the kinematic chain
        :param tip: Tip body of the kinematic chain
        :param target: Desired tip pose relative to the root body
        :param dt: Time step for integration
        :param max_iterations: Maximum number of iterations
        :param translation_velocity: Maximum translation velocity
        :param rotation_velocity: Maximum rotation velocity
        :return: Dictionary mapping DOF names to their computed positions
        """
        target = root._world.transform(target, root)
        qp_problem = QPProblem(
            world=self.world,
            root=root,
            tip=tip,
            target=target,
            dt=dt,
            max_translation_velocity=translation_velocity,
            max_rotation_velocity=rotation_velocity,
        )

        # Initialize solver state
        solver_state = SolverState(
            position=np.array(
                [self.world.state[dof.name].position for dof in qp_problem.active_dofs]
            ),
            passive_position=np.array(
                [self.world.state[dof.name].position for dof in qp_problem.passive_dofs]
            ),
        )

        # Run iterative solver
        final_position = self._solve_iteratively(
            qp_problem, solver_state, dt, max_iterations
        )

        return {dof: final_position[i] for i, dof in enumerate(qp_problem.active_dofs)}

    def _solve_iteratively(
        self,
        qp_problem: QPProblem,
        solver_state: SolverState,
        dt: float,
        max_iterations: int,
    ) -> np.ndarray:
        """
        Tries to solve the inverse kinematics problem iteratively.
        :param qp_problem: Problem definition.
        :param solver_state: Initial state.
        :param dt: Step size per iteration. Unit is seconds.
                    Too large values can lead to instability, too small values can lead to slow convergence.
        :param max_iterations: Maximum number of iterations. A lower dt requires more iterations.
        :return: The final state after max_iterations.
        """
        for self.iteration in range(max_iterations):
            velocity, slack = self._solve_qp_step(qp_problem, solver_state)

            if self._check_convergence(velocity, slack):
                break

            solver_state.update_position(velocity, dt)
        else:
            raise MaxIterationsException(max_iterations)
        return solver_state.position

    def _solve_qp_step(
        self, qp_problem: QPProblem, solver_state: SolverState
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluate the QP matrices at the current state and solve the QP.
        :param qp_problem: Problem definition.
        :param solver_state: Current state
        :return: Velocities for the DOFs, and slack values.
        """

        # Evaluate QP matrices at current state
        qp_matrices = qp_problem.evaluate_at_state(solver_state)

        # Setup constraint sense (equality for last 6 constraints)
        sense = np.zeros(qp_matrices.l.shape, dtype=c_int)
        sense[-6:] = 5  # equality constraints

        # Solve QP
        (xstar, fval, exitflag, info) = daqp.solve(
            qp_matrices.H,
            qp_matrices.g,
            qp_matrices.A,
            qp_matrices.u,
            qp_matrices.l,
            sense,
        )

        if exitflag != 1:
            raise QPSolverException(exitflag)

        return (
            xstar[: len(qp_problem.active_symbols)],
            xstar[len(qp_problem.active_symbols) :],
        )

    def _check_convergence(self, velocity: np.ndarray, slack: np.ndarray) -> bool:
        """
        :param velocity: Current velocity of the DOFs.
        :param slack: Current slack values.
        :return: Whether the solver has converged.
        """
        vel_below_threshold = (
            np.max(np.abs(velocity)) < self._convergence_velocity_tolerance
        )
        slack_below_threshold = (
            np.max(np.abs(slack)) < self._convergence_slack_tolerance
        )
        if vel_below_threshold and slack_below_threshold:
            return True
        if vel_below_threshold and not slack_below_threshold:
            raise UnreachableException(self.iterations)
        return False


@dataclass
class QPProblem:
    """
    Represents a quadratic programming problem for inverse kinematics.
    """

    world: World
    """
    Backreference to semantic digital twin.
    """

    root: KinematicStructureEntity
    """
    Root body of the kinematic chain.
    """

    tip: KinematicStructureEntity
    """
    Tip body of the kinematic chain.
    """

    target: cas.TransformationMatrix
    """
    Desired tip pose relative to the root body.
    """

    dt: float
    """
    Time step for integration.
    """

    max_translation_velocity: float
    max_rotation_velocity: float

    def __post_init__(self):
        # Extract DOFs and setup problem
        (
            self.active_dofs,
            self.passive_dofs,
            self.active_symbols,
            self.passive_symbols,
        ) = self._extract_dofs()
        self._setup_constraints()
        self._setup_weights()
        self._compile_functions()

    def _extract_dofs(
        self,
    ) -> Tuple[
        list[DegreeOfFreedom], list[DegreeOfFreedom], list[cas.Symbol], list[cas.Symbol]
    ]:
        """
        Extract active and passive DOFs from the kinematic chain.
        :return: Active Dofs, Passive Dofs, Active Symbols, Passive Symbols.
        """
        active_dofs_set = set()
        passive_dofs_set = set()
        root_to_common_link, common_link_to_tip = (
            self.world.compute_split_chain_of_connections(self.root, self.tip)
        )
        for connection in root_to_common_link + common_link_to_tip:
            if isinstance(connection, ActiveConnection):
                active_dofs_set.update(connection.active_dofs)
            if isinstance(connection, PassiveConnection):
                passive_dofs_set.update(connection.passive_dofs)

        active_dofs: List[DegreeOfFreedom] = list(
            sorted(active_dofs_set, key=lambda d: str(d.name))
        )
        passive_dofs: List[DegreeOfFreedom] = list(
            sorted(passive_dofs_set, key=lambda d: str(d.name))
        )

        active_symbols = [dof.symbols.position for dof in active_dofs]
        passive_symbols = [dof.symbols.position for dof in passive_dofs]

        return active_dofs, passive_dofs, active_symbols, passive_symbols

    def _setup_constraints(self):
        """Setup all constraints for the QP problem."""
        self.constraint_builder = ConstraintBuilder(
            self.world,
            self.root,
            self.tip,
            self.target,
            self.dt,
            self.max_translation_velocity,
            self.max_rotation_velocity,
        )

        # Box constraints
        self.lower_box_constraints, self.upper_box_constraints = (
            self.constraint_builder.build_box_constraints(self.active_dofs)
        )
        self.box_constraint_matrix = cas.Expression.eye(len(self.lower_box_constraints))

        # Goal constraints
        self.eq_bound_expr, self.neq_matrix = (
            self.constraint_builder.build_goal_constraints(self.active_symbols)
        )

        # Combine constraints
        self.l = cas.Expression.vstack([self.lower_box_constraints, self.eq_bound_expr])
        self.u = cas.Expression.vstack([self.upper_box_constraints, self.eq_bound_expr])
        self.A = cas.Expression.vstack([self.box_constraint_matrix, self.neq_matrix])

    def _setup_weights(self):
        """Setup quadratic and linear weights for the QP problem."""
        dof_weights = [
            0.001 * (1.0 / min(1.0, dof.upper_limits.velocity)) ** 2
            for dof in self.active_dofs
        ]
        slack_weights = [2500 * (1.0 / 0.2) ** 2] * 6

        self.quadratic_weights = cas.Expression(dof_weights + slack_weights)
        self.linear_weights = cas.Expression.zeros(*self.quadratic_weights.shape)

    def _compile_functions(self):
        """Compile all symbolic expressions into functions."""
        symbol_args = [self.active_symbols, self.passive_symbols]

        self.l_f = self.l.compile(symbol_args)
        self.u_f = self.u.compile(symbol_args)
        self.A_f = self.A.compile(symbol_args)
        self.quadratic_weights_f = self.quadratic_weights.compile(symbol_args)
        self.linear_weights_f = self.linear_weights.compile(symbol_args)

    def evaluate_at_state(self, solver_state) -> QPMatrices:
        """Evaluate QP matrices at the current solver state."""
        return QPMatrices(
            H=np.diag(
                self.quadratic_weights_f(
                    solver_state.position, solver_state.passive_position
                )
            ),
            g=self.linear_weights_f(
                solver_state.position, solver_state.passive_position
            ),
            A=self.A_f(solver_state.position, solver_state.passive_position),
            l=self.l_f(solver_state.position, solver_state.passive_position),
            u=self.u_f(solver_state.position, solver_state.passive_position),
        )


@dataclass
class ConstraintBuilder:
    """
    Builds constraints for the inverse kinematics QP problem.
    """

    world: World
    """
    Backreference to semantic digital twin.
    """

    root: KinematicStructureEntity
    """
    Root body of the kinematic chain.
    """

    tip: KinematicStructureEntity
    """
    Tip body of the kinematic chain.
    """

    target: cas.TransformationMatrix
    """
    Desired tip pose relative to the root body.
    """

    dt: float
    """
    Time step for integration.
    """

    max_translation_velocity: float
    max_rotation_velocity: float

    maximum_velocity: float = field(default=1.0, init=False)
    """
    Used to limit the velocity of the DOFs, because the default values defined in the semantic digital twin are sometimes unreasonably high.
    """

    def build_box_constraints(
        self, active_dofs: List[DegreeOfFreedom]
    ) -> Tuple[cas.Expression, cas.Expression]:
        """Build position and velocity limit constraints for DOFs."""
        lower_constraints = []
        upper_constraints = []

        for dof in active_dofs:
            ll = cas.max(-self.maximum_velocity, dof.lower_limits.velocity)
            ul = cas.min(self.maximum_velocity, dof.upper_limits.velocity)

            if dof.has_position_limits():
                ll = cas.max(dof.lower_limits.position - dof.symbols.position, ll)
                ul = cas.min(dof.upper_limits.position - dof.symbols.position, ul)

            lower_constraints.append(ll)
            upper_constraints.append(ul)

        # Add slack variables
        global _large_value
        lower_constraints.extend([-_large_value] * 6)
        upper_constraints.extend([_large_value] * 6)

        return cas.Expression(lower_constraints), cas.Expression(upper_constraints)

    def build_goal_constraints(
        self, active_symbols: List[cas.Symbol]
    ) -> Tuple[cas.Expression, cas.Expression]:
        """Build position and rotation goal constraints."""
        root_T_tip = self.world.compose_forward_kinematics_expression(
            self.root, self.tip
        )

        # Position and rotation errors
        position_state, position_error = self._compute_position_error(root_T_tip)
        rotation_state, rotation_error = self._compute_rotation_error(root_T_tip)

        # Current state and jacobian
        current_expr = cas.Expression.vstack([position_state, rotation_state])
        eq_bound_expr = cas.Expression.vstack([position_error, rotation_error])

        J = current_expr.jacobian(active_symbols)
        neq_matrix = cas.Expression.hstack(
            [J * self.dt, cas.Expression.eye(6) * self.dt]
        )

        return eq_bound_expr, neq_matrix

    def _compute_position_error(
        self, root_T_tip: cas.TransformationMatrix
    ) -> Tuple[cas.Expression, cas.Expression]:
        """
        Compute position error with velocity limits.
        :param root_T_tip: Forward kinematics expression.
        :return: Expression describing the position, and the error vector.
        """
        root_P_tip = root_T_tip.to_position()
        root_T_tip_goal = cas.TransformationMatrix(self.target)
        root_P_tip_goal = root_T_tip_goal.to_position()

        translation_cap = self.max_translation_velocity * self.dt
        position_error = root_P_tip_goal[:3] - root_P_tip[:3]

        for i in range(3):
            position_error[i] = cas.limit(
                position_error[i], -translation_cap, translation_cap
            )

        return root_P_tip[:3], position_error

    def _compute_rotation_error(
        self, root_T_tip: cas.TransformationMatrix
    ) -> Tuple[cas.Expression, cas.Expression]:
        """
        Compute rotation error with velocity limits.
        :param root_T_tip: Forward kinematics expression.
        :return: Expression describing the rotation, and the error vector.
        """
        rotation_cap = self.max_rotation_velocity * self.dt

        hack = cas.RotationMatrix.from_axis_angle(cas.Vector3.Z(), -0.0001)
        root_R_tip = root_T_tip.to_rotation_matrix().dot(hack)
        q_actual = cas.TransformationMatrix(self.target).to_quaternion()
        q_goal = root_R_tip.to_quaternion()
        q_goal = cas.if_less(q_goal.dot(q_actual), 0, -q_goal, q_goal)
        q_error = q_actual.diff(q_goal)

        rotation_error = -q_error
        for i in range(3):
            rotation_error[i] = cas.limit(
                rotation_error[i], -rotation_cap, rotation_cap
            )

        return q_error[:3], rotation_error[:3]


class SolverState:
    """
    Represents the state of the IK solver during iteration.
    """

    def __init__(self, position: np.ndarray, passive_position: np.ndarray):
        self.position = position
        self.passive_position = passive_position
        self.positions_history = []
        self.velocities_history = []

    def update_position(self, velocity: np.ndarray, dt: float):
        self.positions_history.append(self.position.copy())
        self.velocities_history.append(velocity.copy())
        self.position += velocity * dt


@dataclass
class QPMatrices:
    """
    Container for QP problem matrices at a specific state.
    min_{x} x^T H x + g^T x
    subject to
        l <= Ax <= u

    Find an x that minimizes the cost function, while satisfying the constraints.
    H (for Hessian) describes a quadratic cost function and g (for gradient) a linear one.
    Ax is the constraint space velocity, e.g. if A is the Jacobian, then Ax is the Cartesian velocity.
    l and u are the lower and upper bounds of the constraint space, e.g., translational or rotational velocity
    if l == u, we have equality constraints.
    """

    H: np.ndarray
    g: np.ndarray
    A: np.ndarray
    l: np.ndarray
    u: np.ndarray



from semantic_digital_twin.testing import two_arm_robot_world
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.world_entity import Body


def test_simple_two_arm_robot(two_arm_robot_world: World):
    for c in two_arm_robot_world.connections:
        assert isinstance(c.origin_expression.reference_frame, Body), c.name
        assert isinstance(c.origin_expression.child_frame, Body), c.name
    for b in two_arm_robot_world.bodies:
        for c in b.collision:
            assert isinstance(c.origin.reference_frame, Body), b.name
            assert c.origin.child_frame is None, b.name
        for v in b.visual:
            assert isinstance(v.origin.reference_frame, Body), b.name
            assert v.origin.child_frame is None, b.name
